#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include <GL/glew.h>
#include <GL/glu.h>

#include <SDL.h>
#include <SDL_opengl.h>
#include <SDL_image.h>

#include "utils_graphics.h"
#include "shaders.h"
#include "render_text.h"



#define SCREEN_WIDTH        640
#define SCREEN_HEIGHT       480

#define POSITION            0

#define MAX_SNOW_FLAKES     6
#define MAX_VERTEX          20000

// rozmiary bryły obcinania
GLfloat left = -2.0f;
GLfloat right = 2.0f;
GLfloat bottom = -2.0f;
GLfloat top = 2.0f;
GLfloat near = 3.0f;
GLfloat far = 7.0f;

GLfloat rotate_x = 0.0f;
GLfloat rotate_y = 0.0f;


GLuint vertex_shader;
GLuint fragment_shader;
GLuint shader_program;

GLuint vertex_array;
GLuint vertex_buffer;
GLuint texture;


GLfloat position[MAX_VERTEX * 4];
int sprite_count = 1000;
GLfloat point_size = 10.0f;

float projection_matrix[16];
float modelview_matrix[16];
float mvp_matrix[16];
float aux_matrix1[16];
float aux_matrix2[16];
float texture_matrix[16];

SDL_Window* g_window = NULL;
SDL_GLContext g_context;
SDL_Surface *image2d;
unsigned int last_time;


int init();
int initGL();
int set_shaders();
int init_buffers();
void cleanup();
void update();
void render();
int check_gl_error();

char *text_file_read(char *fname);



int init() {
    // Initialize SDL
    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
        printf("Błąd podczas inicjalizacji silnika SDL2: %s\n", SDL_GetError());
        return 0;
    }

    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 1);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);

    g_window = SDL_CreateWindow("Śnieg", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
                                SCREEN_WIDTH, SCREEN_HEIGHT, SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN);
    if (g_window == NULL) {
        printf("Błąd podczas tworzenia okna: %s\n", SDL_GetError());
        return 0;
    }

    g_context = SDL_GL_CreateContext(g_window);
    if (g_context == NULL) {
        printf("Błąd podczas tworzenia kontekstu OpenGL: %s\n", SDL_GetError());
        return 0;
    }

    if (SDL_GL_SetSwapInterval(1) < 0) {
        printf( "Błąd podczas ustawiania VSync: %s\n", SDL_GetError());
    }

    int imgFlags = IMG_INIT_PNG | IMG_INIT_JPG;
    if (!(IMG_Init(imgFlags) & imgFlags)) {
        printf("Błąd podczas inicjalizacji biblioteki SDL_image: %s\n", IMG_GetError());
        return 0;
    }

    if (!initGL()) {
        printf("Błąd podczas inicjalizacji OpenGL!\n");
        return 0;
    }

    if (!init_draw_text()) {
        printf("Error in function init_draw_text\n");
        return 0;
    }

    last_time = SDL_GetTicks();
    return 1;
}

int initGL() {
    GLenum err = GL_NO_ERROR;
    unsigned int width = SCREEN_WIDTH;
    unsigned int height = SCREEN_HEIGHT;
    int i;

    glewExperimental = GL_TRUE;
    err = glewInit();
    while (check_gl_error() == GL_INVALID_ENUM) {};
    if (err != GLEW_OK) {
        printf("glewInit failed: %s\n", glewGetErrorString(err));
        cleanup();
        return 0;
    }

    glClearColor(0.0f, 0.0f, 0.6f, 1.0f);

    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_2D_ARRAY, texture);

    const char *file_names[MAX_SNOW_FLAKES] = {
        "Snowflake1.jpg",
        "Snowflake2.jpg",
        "Snowflake3.jpg",
        "Snowflake4.jpg",
        "Snowflake5.jpg",
        "Snowflake6.jpg"
    };
    for (i = 0; i < MAX_SNOW_FLAKES; i++) {
        image2d = IMG_Load(file_names[i]);
        if (image2d == NULL) {
            printf("Błąd podczas ładowania pliku tekstury %s: %s\n", file_names[i], SDL_GetError());
            return 0;
        }

        glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
        printf("[%s]->BytesPerPixel: %d, w: %d  h: %d\n", file_names[i], image2d->format->BytesPerPixel, image2d->w, image2d->h);
        if (i == 0) {
            // definiowanie pustego obrazu tekstury, po to żeby móc później do niego przypisać kolejne warstwy
            glTexImage3D(GL_TEXTURE_2D_ARRAY, 0, GL_RGB, image2d->w, image2d->h, MAX_SNOW_FLAKES, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
            //glTexStorage3D(GL_TEXTURE_3D, 7, GL_RGBA, image2d->w, image2d->h, layers);
        }

        // definiowanie obrazu warstwy tekstury
        glTexSubImage3D(GL_TEXTURE_2D_ARRAY, 0, 0, 0, i, image2d->w, image2d->h, 1, GL_RGB, GL_UNSIGNED_BYTE, image2d->pixels);

        SDL_FreeSurface(image2d);
        image2d = NULL;
    }
    glGenerateMipmap(GL_TEXTURE_2D_ARRAY);
    glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glBindTexture(GL_TEXTURE_2D_ARRAY, 0);

    glViewport(0, 0, width, height);
    if (width < height && width > 0) {
        get_frustum_matrix(projection_matrix, left, right, bottom*height/width, top*height/width, near, far);
    } else {
        if (width >= height && height > 0) {
            get_frustum_matrix(projection_matrix, left*width/height, right*width/height, bottom, top, near, far);
        } else {
            get_frustum_matrix(projection_matrix, left, right, bottom, top, near, far);
        }
    }
    //dump_matrix4x4(projection_matrix);

    if (!set_shaders()) {
        cleanup();
        return 0;
    }
    if (!init_buffers()) {
        cleanup();
        return 0;
    }

    glEnable(GL_PROGRAM_POINT_SIZE);
    glEnable(GL_POINT_SPRITE);
    glEnable(GL_DEPTH_TEST);

    return 1;
}

int set_shaders() {
    char *shader_source = NULL;
    GLint status;

    shader_source = text_file_read("snow_vs.glsl");
    if (!shader_source) {
        printf("Nie można odczytać shadera wierzchołków\n");
        return 0;
    }
    //printf("%s\n---\n", shader_source);
    vertex_shader = glCreateShader(GL_VERTEX_SHADER);
    //printf("vertex_shader: %d\n", vertex_shader);
    glShaderSource(vertex_shader,  1, (const char* const*) &shader_source, NULL);
    free(shader_source);
    glCompileShader(vertex_shader);
    glGetShaderiv(vertex_shader, GL_COMPILE_STATUS, &status);
    if (status != GL_TRUE) {
        printf("Niepoprawna kompilacja shadera wierzchołków\n");
        print_shader_info_log(vertex_shader);
        return 0;
    }


    shader_source = text_file_read("snow_fs.glsl");
    if (!shader_source) {
        printf("Nie można odczytać shadera fragmentów\n");
        return 0;
    }
    //printf("%s\n---\n", shader_source);
    fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);
    //printf("fragment_shader: %d\n", fragment_shader);
    glShaderSource(fragment_shader,  1, (const char* const*) &shader_source, NULL);
    free(shader_source);
    glCompileShader(fragment_shader);
    glGetShaderiv(fragment_shader, GL_COMPILE_STATUS, &status);
    if (status != GL_TRUE) {
        printf("Niepoprawna kompilacja shadera fragmentów\n");
        print_shader_info_log(fragment_shader);
        return 0;
    }

    shader_program = glCreateProgram();
    glAttachShader(shader_program, vertex_shader);
    glAttachShader(shader_program, fragment_shader);
    glLinkProgram(shader_program);
    glGetProgramiv(shader_program, GL_LINK_STATUS, &status);
    if (status == GL_FALSE) {
        printf("Niepoprawna konsolidacja programu\n");
        print_program_info_log(shader_program);
        return 0;
    }
    glValidateProgram(shader_program);
    glGetProgramiv(shader_program, GL_VALIDATE_STATUS, &status);
    if (status == GL_FALSE) {
        printf("Niepoprawna walidacja programu\n");
        print_program_info_log(shader_program);
        return 0;
    }

    glUseProgram(shader_program);

    return 1;
}


int init_buffers() {
    int i;
    for (i = 0; i < MAX_VERTEX; i++) {
        position[i*4 + 0] = 2.0f*(rand() / ( float )RAND_MAX) * (right - left) + 2.0f*left;
        position[i*4 + 1] = 2.0f*(rand() / ( float )RAND_MAX) * (top - bottom) + 2.0f*top;
        position[i*4 + 2] = (rand() / ( float )RAND_MAX) * (far - near) - near;

        position[i*4 + 3] = (rand() / ( float )RAND_MAX) * MAX_SNOW_FLAKES;
    }

    glGenVertexArrays(1, &vertex_array);
    glBindVertexArray(vertex_array);

    glGenBuffers(1, &vertex_buffer);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer);
    glBufferData(GL_ARRAY_BUFFER, sizeof(position), position, GL_DYNAMIC_DRAW);
    glVertexAttribPointer(POSITION, 4, GL_FLOAT, GL_FALSE, 0, NULL);
    glEnableVertexAttribArray(POSITION);

    glBindVertexArray(0);

    return 1;
}


void cleanup() {
    cleanup_draw_text();

    if (vertex_buffer) {
        glDeleteBuffers(1, &vertex_buffer);
    }
    if (vertex_array) {
        glDeleteVertexArrays(1, &vertex_array);
    }

    if (shader_program && fragment_shader) {
        glDetachShader(shader_program, fragment_shader);
    }
    if (shader_program && vertex_shader) {
        glDetachShader(shader_program, vertex_shader);
    }
    if (fragment_shader) {
        glDeleteShader(fragment_shader);
    }
    if (vertex_shader) {
        glDeleteShader(vertex_shader);
    }
    if (shader_program) {
        glDeleteProgram(shader_program);
    }

    if (texture) {
        glDeleteTextures(1, &texture);
    }

    if (g_context) {
        SDL_GL_DeleteContext(g_context);
        g_context = NULL;
    }

    if (g_window) {
        SDL_DestroyWindow(g_window);
        g_window = NULL;
    }

    IMG_Quit();
    SDL_Quit();
}

void update() {
    unsigned int time_now = SDL_GetTicks();
    if (last_time != 0) {
        unsigned long elapsed = time_now - last_time;
    }
    last_time = time_now;

    for (int i = 0; i < sprite_count; i++) {
        position[i*4 + 1] -= 0.005f + rand() / (float)RAND_MAX * 0.01f;
        position[i*4 + 0] += 0.001f * sin(position[i*3 + 1]);
        position[i*4 + 2] += 0.001f * cos(position[i*3 + 1]);

        if (position[i*4 + 1] < 2.0*bottom) {
            position[i*4 + 0] = 2.0f*(rand() / (float)RAND_MAX) * (right - left) + 2.0f*left;
            position[i*4 + 1] = 2.0f*(rand() / (float)RAND_MAX) * (top - bottom) + 2.0f*top;
            position[i*4 + 2] = (rand() / (float)RAND_MAX) * (far - near) - near;

            position[i*4 + 3] = (rand() / (float)RAND_MAX) * MAX_SNOW_FLAKES;
        }
    }
    glBindVertexArray(vertex_array);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer);
    glBufferSubData(GL_ARRAY_BUFFER, 0, 4 * sprite_count * sizeof(GLfloat), position);
    glBindVertexArray(0);
}

void render() {
    char buf[256];

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    get_identity_matrix(modelview_matrix);

    get_translation_matrix4x4(aux_matrix1, 0.0f, 0.0f, -(near+far)/2.0f);
    mult_matrices(aux_matrix2, modelview_matrix, aux_matrix1, 4);

    get_rotation_x_matrix4x4(aux_matrix1, rotate_x);
    mult_matrices(modelview_matrix, aux_matrix2, aux_matrix1, 4);
    get_rotation_y_matrix4x4(aux_matrix1, rotate_y);
    mult_matrices(aux_matrix2, modelview_matrix, aux_matrix1, 4);
    memcpy(modelview_matrix, aux_matrix2, sizeof(float) * 16);

    mult_matrices(aux_matrix1, projection_matrix, modelview_matrix, 4);
    transpose_matrix(mvp_matrix, aux_matrix1, 4);



    glUseProgram(shader_program);
    glBindVertexArray(vertex_array);
    glBindTexture(GL_TEXTURE_2D_ARRAY, texture);

    glUniformMatrix4fv(glGetUniformLocation(shader_program, "modelViewProjectionMatrix"), 1, GL_FALSE, mvp_matrix);
    glUniform1i(glGetUniformLocation(shader_program, "tex"), 0);

    // narysowanie danych zawartych w tablicach wierzchołków
    glUniform1f(glGetUniformLocation(shader_program, "pointSize"), point_size);
    glDrawArrays(GL_POINTS, 0, sprite_count);


    glBindTexture(GL_TEXTURE_2D_ARRAY, 0);
    glUseProgram(0);
    glBindVertexArray(0);

    glDisable(GL_DEPTH_TEST);
    sprintf(buf, "liczba czastek: %d", sprite_count);
    draw_text(6.0f, 20.0f, buf, 11.0f, 1.0f, 1.0f, 1.0f, 1.0f);
    glEnable(GL_DEPTH_TEST);
}



int main(int argc, char* args[]) {
    int quit;
    SDL_Event e;

    srand(time(NULL));

    if (!init()) {
        printf("Błąd inicjalizacji!\n");
        cleanup();
        return 1;
    }

    quit = 0;
    last_time = SDL_GetTicks();
    while (!quit) {
        while (SDL_PollEvent(&e) != 0) {
            if (e.type == SDL_QUIT) {
                quit = 1;
                break;
            } else if (e.type == SDL_KEYDOWN) {
                switch (e.key.keysym.sym) {
                    case SDLK_LEFT:
                        rotate_y -= 0.2f;
                        break;
                    case SDLK_RIGHT:
                        rotate_y += 0.2f;
                        break;
                    case SDLK_UP:
                        rotate_x += 0.2f;
                        break;
                    case SDLK_DOWN:
                        rotate_x -= 0.2f;
                        break;
                    case SDLK_PAGEUP:
                        point_size += 0.2f;
                        break;
                    case SDLK_PAGEDOWN:
                        if (point_size > 0.2f) {
                            point_size -= 0.2f;
                        }
                        break;
                    case SDLK_KP_PLUS:
                        if (sprite_count + 1000 <= 20000) {
                            sprite_count += 1000;
                        }
                        break;
                    case SDLK_KP_MINUS:
                        if (sprite_count - 1000 > 0) {
                            sprite_count -= 1000;
                        }
                        break;
                };
            }
        }

        update();
        render();

        SDL_GL_SwapWindow(g_window);
    }

    cleanup();

    return 0;
}



char *text_file_read(char *fname) {
    FILE *fp;
    char *content = NULL;
    int count = 0;

    if (fname != NULL) {
        fp = fopen(fname, "rt");

        if (fp != NULL) {
            fseek(fp, 0, SEEK_END);
            count = ftell(fp);
            rewind(fp);

            if (count > 0) {
                content = (char*)malloc(sizeof(char) * (count+1));
                count = fread(content,sizeof(char),count,fp);
                content[count] = '\0';
            }
            fclose(fp);
        }
    }
    return content;
}


int check_gl_error() {
    GLenum error;

    error = glGetError();
    switch (error) {
        case GL_CONTEXT_LOST:
            printf("GL_CONTEXT_LOST [%d]: %s\n", error, gluErrorString(error));
            return GL_CONTEXT_LOST;
        case GL_INVALID_ENUM:
            printf("GL_INVALID_ENUM [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_ENUM;
        case GL_INVALID_VALUE:
            printf("GL_INVALID_VALUE [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_VALUE;
        case GL_INVALID_OPERATION:
            printf("GL_INVALID_OPERATION [%d]: %s\n", error, gluErrorString(error));
            return 1;
        case GL_INVALID_FRAMEBUFFER_OPERATION:
            printf("GL_INVALID_FRAMEBUFFER_OPERATION [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_FRAMEBUFFER_OPERATION;
        case GL_OUT_OF_MEMORY:
            printf("GL_OUT_OF_MEMORY [%d]: %s\n", error, gluErrorString(error));
            return 1;
        case GL_STACK_OVERFLOW:
            printf("GL_STACK_OVERFLOW [%d]: %s\n", error, gluErrorString(error));
            return GL_STACK_OVERFLOW;
        case GL_STACK_UNDERFLOW:
            printf("GL_STACK_UNDERFLOW [%d]: %s\n", error, gluErrorString(error));
            return GL_STACK_UNDERFLOW;
        case GL_NO_ERROR:
            break;
    }

    return GL_NO_ERROR;
}
