#include "utils_graphics.h"
#include <stdio.h>
#include <math.h>

void dump_matrix4x4(float *matrix) {
    fprintf(stdout, "%.2f\t%.2f\t%.2f\t%.2f\n"
                    "%.2f\t%.2f\t%.2f\t%.2f\n"
                    "%.2f\t%.2f\t%.2f\t%.2f\n"
                    "%.2f\t%.2f\t%.2f\t%.2f\n",
                    matrix[0], matrix[1], matrix[2], matrix[3],
                    matrix[4], matrix[5], matrix[6], matrix[7],
                    matrix[8], matrix[9], matrix[10], matrix[11],
                    matrix[12], matrix[13], matrix[14], matrix[15]);
}

void dump_vector4(float *vector) {
    fprintf(stdout, "%.2f\n"
                    "%.2f\n"
                    "%.2f\n"
                    "%.2f\n",
                    vector[0], vector[1], vector[2], vector[3]);
}

void mult_matrices(float *output_matrix, float *matrix1, float *matrix2, int matrix_size) {
    int i, j, k;
    float val;

    for (i = 0; i < matrix_size; i++) {
        for (j = 0; j < matrix_size; j++) {
            val = 0.0f;
            for (k = 0; k < matrix_size; k++) {
                val += matrix1[j * matrix_size + k] * matrix2[i + k * matrix_size];
            }
            output_matrix[i + (j * matrix_size)] = val;
        }
    }
}

void transform_vector(float *output_vector, float *input_matrix, float *input_vector, int vec_matrix_size) {
    int i,j;
    float val;

    for (i = 0; i < vec_matrix_size; i++) {
        val = 0;
        for (j = 0; j < vec_matrix_size; j++) {
            val += input_matrix[i*vec_matrix_size + j] * input_vector[j];
        }
        output_vector[i] = val;
    }
}

void get_rotation_x_matrix4x4(float *output_matrix, float angle) {
    output_matrix[0] = 1;
    output_matrix[1] = 0;
    output_matrix[2] = 0;
    output_matrix[3] = 0;

    output_matrix[4] = 0;
    output_matrix[5] = cos(angle);
    output_matrix[6] = -sin(angle);
    output_matrix[7] = 0;

    output_matrix[8] = 0;
    output_matrix[9] = sin(angle);
    output_matrix[10] = cos(angle);
    output_matrix[11] = 0;

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void get_rotation_y_matrix4x4(float *output_matrix, float angle) {
    output_matrix[0] = cos(angle);
    output_matrix[1] = 0;
    output_matrix[2] = sin(angle);
    output_matrix[3] = 0;

    output_matrix[4] = 0;
    output_matrix[5] = 1;
    output_matrix[6] = 0;
    output_matrix[7] = 0;

    output_matrix[8] = -sin(angle);
    output_matrix[9] = 0;
    output_matrix[10] = cos(angle);
    output_matrix[11] = 0;

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void get_rotation_z_matrix4x4(float *output_matrix, float angle) {
    output_matrix[0] = cos(angle);
    output_matrix[1] = -sin(angle);
    output_matrix[2] = 0;
    output_matrix[3] = 0;

    output_matrix[4] = sin(angle);
    output_matrix[5] = cos(angle);
    output_matrix[6] = 0;
    output_matrix[7] = 0;

    output_matrix[8] = 0;
    output_matrix[9] = 0;
    output_matrix[10] = 1;
    output_matrix[11] = 0;

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void get_translation_matrix4x4(float *output_matrix, float x, float y, float z) {
    output_matrix[0] = 1;
    output_matrix[1] = 0;
    output_matrix[2] = 0;
    output_matrix[3] = x;

    output_matrix[4] = 0;
    output_matrix[5] = 1;
    output_matrix[6] = 0;
    output_matrix[7] = y;

    output_matrix[8] = 0;
    output_matrix[9] = 0;
    output_matrix[10] = 1;
    output_matrix[11] = z;

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void get_scale_matrix4x4(float *output_matrix, float scale_x, float scale_y, float scale_z) {
    output_matrix[0] = scale_x;
    output_matrix[1] = 0;
    output_matrix[2] = 0;
    output_matrix[3] = 0;

    output_matrix[4] = 0;
    output_matrix[5] = scale_y;
    output_matrix[6] = 0;
    output_matrix[7] = 0;

    output_matrix[8] = 0;
    output_matrix[9] = 0;
    output_matrix[10] = scale_z;
    output_matrix[11] = 0;

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void transpose_matrix(float *output_matrix, float *input_matrix, int matrix_size) {
    int i, j;

    for (i = 0; i < matrix_size; i++) {
        for (j = 0; j < matrix_size; j++) {
            output_matrix[i*matrix_size + j] = input_matrix[i + j*matrix_size];
        }
    }
}

void get_identity_matrix(float *output_matrix) {
    output_matrix[0] = 1;
    output_matrix[1] = 0;
    output_matrix[2] = 0;
    output_matrix[3] = 0;

    output_matrix[4] = 0;
    output_matrix[5] = 1;
    output_matrix[6] = 0;
    output_matrix[7] = 0;

    output_matrix[8] = 0;
    output_matrix[9] = 0;
    output_matrix[10] = 1;
    output_matrix[11] = 0;

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void get_ortho_matrix(float *output_matrix, float left, float right, float bottom, float top, float near, float far) {
    output_matrix[0] = 2 / (right - left);
    output_matrix[1] = 0;
    output_matrix[2] = 0;
    output_matrix[3] = -(right + left) / (right - left);

    output_matrix[4] = 0;
    output_matrix[5] = 2 / (top - bottom);
    output_matrix[6] = 0;
    output_matrix[7] = -(top + bottom) / (top - bottom);

    output_matrix[8] = 0;
    output_matrix[9] = 0;
    output_matrix[10] = -2 / (far - near);
    output_matrix[11] = -(far + near) / (far - near);

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void get_frustum_matrix(float *output_matrix, float left, float right, float bottom, float top, float near, float far) {
    output_matrix[0] = 2 * near / (right - left);
    output_matrix[1] = 0;
    output_matrix[2] = (right + left) / (right - left);
    output_matrix[3] = 0;

    output_matrix[4] = 0;
    output_matrix[5] = 2 * near / (top - bottom);
    output_matrix[6] = (top + bottom) / (top - bottom);
    output_matrix[7] = 0;

    output_matrix[8] = 0;
    output_matrix[9] = 0;
    output_matrix[10] = -(far + near) / (far - near);
    output_matrix[11] = -2 * far * near / (far - near);

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = -1;
    output_matrix[15] = 0;
}
